import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime
from pprint import pprint

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq

from ds_capability import FeatureEngineer, FeatureSelect
from ds_capability.components.commons import Commons
from ds_core.properties.property_manager import PropertyManager
from ds_capability.intent.feature_select_intent import FeatureSelectIntent

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureEngineererTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_auto_reinstate_nulls(self):
        c = pa.array(['a', 'y', 'f', '', 'a', '', 'p', ])
        tbl = pa.table([c], names=['string'])
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        result = tools.auto_reinstate_nulls(tbl)
        self.assertEqual(2, result.column('string').null_count)

    def test_auto_drop_noise(self):
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(1000, extend=True)
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        self.assertEqual(19, tbl.num_columns)
        result = tools.auto_drop_noise(tbl)
        self.assertEqual(13, result.num_columns)
        print(result.column_names)

    def test_auto_drop_correlated(self):
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(1000, extend=True)
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        self.assertEqual(19, tbl.num_columns)
        self.assertIn('dup_num', tbl.column_names)
        result = tools.auto_drop_correlated(tbl)
        self.assertEqual(18, result.num_columns)
        self.assertNotIn('dup_num', result.column_names)

    def test_aggrigate(self):
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(5, extend=False)
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        result = tools.auto_aggregate(tbl, action='count', regex=['cat', 'string'], to_header='agg', drop_aggregated=True)
        # print(result.column('agg').to_pylist())
        # print(result.column_names)
        self.assertEqual([2, 2, 2, 2, 2], result.column('agg').to_pylist())
        self.assertEqual(['num', 'int', 'bool', 'date', 'agg'], result.column_names)

    def test_auto_projection(self):
        tbl = FeatureEngineer.from_memory().tools.get_noise(size=1000, num_columns=5)
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        result = tools.auto_projection(tbl, n_components=2)
        control = ['pca_A', 'pca_B']
        self.assertEqual(control, result.column_names)
        self.assertEqual((1000, 2), result.shape)

    def test_auto_append_tables(self):
        tbl = FeatureEngineer.from_memory().tools.get_noise(size=1000, num_columns=5)
        fs = FeatureSelect.from_memory()
        tools: FeatureSelectIntent = fs.tools
        result = tools.auto_append_tables(tbl, None)
        self.assertTrue(tbl.equals(result))
        result = tools.auto_append_tables(tbl, tbl)
        self.assertTrue(tbl.equals(result))
        tbl_types = FeatureEngineer.from_memory().tools.get_synthetic_data_types(1200)
        result = tools.auto_append_tables(tbl, tbl_types, headers=['A', 'B', 'C'], other_headers=['num', 'int'])
        self.assertEqual((1000, 5), result.shape)
        self.assertCountEqual(['A', 'B', 'C', 'num', 'int'], result.column_names)
        tbl_types = FeatureEngineer.from_memory().tools.get_synthetic_data_types(400)
        result = tools.auto_append_tables(tbl, tbl_types, headers=['A', 'B', 'C'], other_headers=['num', 'int'])
        self.assertEqual((1000, 5), result.shape)
        self.assertCountEqual(['A', 'B', 'C', 'num', 'int'], result.column_names)

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def get_table():
    num = pa.array([1.0, None, 5.0, -0.46421, 3.5, 7.233, -2], pa.float64())
    val = pa.array([1, 2, 3, 4, 5, 6, 7], pa.int64())
    date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, None, "2023-01-02 05:23:50", None, None],
                       format='%Y-%m-%d %H:%M:%S', unit='ns')
    text = pa.array(["Blue", "Green", None, 'Red', 'Orange', 'Yellow', 'Pink'], pa.string())
    binary = pa.array([True, True, None, False, False, True, False], pa.bool_())
    cat = pa.array([None, 'M', 'F', 'M', 'F', 'M', 'M'], pa.string()).dictionary_encode()
    return pa.table([num, val, date, text, binary, cat], names=['num', 'int', 'date', 'text', 'bool', 'cat'])


def pm_view(capability: str, task: str, section: str = None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability, {}).get(task, {})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result


if __name__ == '__main__':
    unittest.main()



from typing import Any

from ds_core.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler
from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory
import pyarrow as pa

class PostgresSourceHandler(AbstractSourceHandler):
    """ A Postgres Source Handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the source_contract dictionary """
        # required module import
        self.psycopg2 = HandlerFactory.get_module('psycopg2')
        super().__init__(connector_contract)
        connector_type = self.connector_contract.schema
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))
        self._file_state = 0
        self._changed_flag = True

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['postgresql', 'postgres']

    def exists(self) -> bool:
        return True

    def has_changed(self) -> bool:
        """ returns if the file has been modified"""
        # TODO: Add in change logic here
        state = None
        if state != self._file_state:
            self._changed_flag = True
            self._file_state = state
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed

    def load_canonical(self, **kwargs) -> dict:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        conn = None
        cur = self.get_cursor()
        _kwargs = {**self.connector_contract.query, **kwargs}
        table = _kwargs.pop('table', 'hadron_table')
        sql_query = _kwargs.pop('sql_query', f"SELECT * FROM {table};")
        try:
            cur.execute(sql_query)
            col_names = [desc[0] for desc in cur.description]
            rows = cur.fetchall()
            return pa.table([rows], names=[col_names])
        except (Exception, self.psycopg2.DatabaseError) as error:
            print(error)
        finally:
            if conn is not None:
                conn.close()
                print('Database connection closed.')

    def get_cursor(self):
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        _kwargs = self.connector_contract.query
        config_name = _kwargs.pop('config_name', None)
        section = _kwargs.pop('config_section', 'postgresql')
        try:
            config = self.connector_contract.parse_ini(config_name=config_name, section=section)
        except ValueError:
            config = {}
        database = config.pop('database', self.connector_contract.path[1:])
        host = config.pop('host', self.connector_contract.hostname)
        port = config.pop('post', self.connector_contract.port)
        user = config.pop('user', self.connector_contract.username)
        password = config.pop('password', self.connector_contract.password)
        try:
            conn = self.psycopg2.connect(database=database, host=host, port=port, user=user, password=password)
            return conn.cursor()
        except (Exception, self.psycopg2.DatabaseError) as error:
            raise ConnectionError(error)


class PostgresPersistHandler(PostgresSourceHandler, AbstractPersistHandler):
    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        return self.backup_canonical(canonical=canonical, uri=self.connector_contract.uri, **kwargs)

    def backup_canonical(self, canonical: Any, uri: str, **kwargs) -> bool:
        conn = None
        cur = self.get_cursor()
        _kwargs = {**self.connector_contract.query, **kwargs}
        table = _kwargs.pop('table', 'hadron_table')
        sql_query = _kwargs.pop('sql_query', f"CREATE OR REPLACE TABLE {table} AS SELECT * FROM canonical;")
        try:
            cur.execute(sql_query)
            return
        except (Exception, self.psycopg2.DatabaseError) as error:
            print(error)
        finally:
            if conn is not None:
                conn.close()
                print('Database connection closed.')

    def remove_canonical(self, **kwargs) -> bool:
        pass



import os
import pyarrow as pa
import pyarrow.compute as pc
from ds_core.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler
from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory

class DuckdbSourceHandler(AbstractSourceHandler):
    """ A DuckDB source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the source_contract dictionary """
        # required module import
        self.duckdb = HandlerFactory.get_module('duckdb')
        super().__init__(connector_contract)
        # connection
        self.connection = self.duckdb.connect(':default:')
        # address
        self.address = connector_contract.address
        if self.address.endswith('duckdb:'):
            self.address = ':default:'
        # remote table
        _kwargs = self.connector_contract.query
        table = _kwargs.pop('table', 'hadron_table')
        if table.startswith('s3//'):
            region_name = _kwargs.pop('region_name', os.environ.get('AWS_DEFAULT_REGION'))
            aws_access_key_id = _kwargs.pop('aws_access_key_id', os.environ.get('AWS_ACCESS_KEY_ID'))
            aws_secret_access_key = _kwargs.pop('aws_secret_access_key', os.environ.get('AWS_SECRET_ACCESS_KEY'))
            aws_session_token = _kwargs.pop('aws_session_token', os.environ.get('AWS_SESSION_TOKEN'))
            self.connection.execute(f"""
                INSTALL httpfs;
                LOAD httpfs;
                SET s3_region='{region_name}';
                SET s3_access_key_id='{aws_access_key_id}';
                SET s3_secret_access_key='{aws_secret_access_key}';
                SET s3_session_token='{aws_session_token}';
            """)
        elif table.startswith('http//'):
            self.connection.execute(f"""
                INSTALL httpfs;
                LOAD httpfs;
             """)
        self.first_row = None
        self._changed_flag = True

    def supported_types(self) -> list:
        return ['parquet']

    def exists(self) -> bool:
        _kwargs = self.connector_contract.query
        table = _kwargs.pop('table', 'hadron_table')
        result = self.connection.execute("CALL duckdb_tables()").arrow()
        return pc.is_in(table, result.column('table_name')).as_py()

    def has_changed(self) -> bool:
        _kwargs = self.connector_contract.query
        table = _kwargs.pop('table', 'hadron_table')
        result = self.connection.execute(f"SELECT * FROM {table} LIMIT 1;").arrow()
        if self.first_row.equals(result):
            return False
        return True

    def reset_changed(self, changed: bool=None):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed
        if changed:
            _kwargs = self.connector_contract.query
            table = _kwargs.pop('table', 'hadron_table')
            result = self.connection.execute(f"SELECT * FROM {table} LIMIT 1;").arrow()
            self.first_row = result

    def load_canonical(self, **kwargs) -> pa.Table:
        _kwargs = {**self.connector_contract.query, **kwargs}
        table = _kwargs.pop('table', 'hadron_table')
        if table.startswith("s3://"):
            return self.connection.execute(f"SELECT * FROM read_parquet('{table}')").arrow()
        elif table.startswith('https://'):
            return self.connection.execute(f"SELECT * FROM read_parquet('{table}')").arrow()
        query = _kwargs.pop('sql_query', f"SELECT * FROM {table};")
        query = query.replace('@', table)
        return self.connection.execute(query).arrow()


class DuckdbPersistHandler(DuckdbSourceHandler, AbstractPersistHandler):

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _kwargs = {**self.connector_contract.query, **kwargs}
        table = _kwargs.pop('table', 'hadron_table')
        if table.startswith("s3://"):
            self.connection.execute(f"COPY tbl TO '{table}'")
            return
        elif table.startswith('http//'):
            raise NotImplementedError("Not supported. 'https' is currently read only")
        query = _kwargs.pop('sql_query', f"CREATE OR REPLACE TABLE {table} AS SELECT * FROM canonical;")
        self.connection.execute(query)
        # reset has changed
        self.reset_changed(True)
        return

    def remove_canonical(self, **kwargs) -> bool:
        _kwargs = {**self.connector_contract.query, **kwargs}
        table = _kwargs.pop('table', 'hadron_table')
        self.connection.execute(f"DROP TABLE IF EXISTS {table}")
        return True

    def backup_canonical(self, canonical: pa.Table, uri: str, **kwargs) -> bool:
        pass




import datetime
import time
import pandas as pd
import pyarrow as pa
from ds_core.handlers.abstract_handlers import ConnectorContract
from ds_core.components.abstract_component import AbstractComponent

from ds_capability.components.commons import Commons
from ds_capability.managers.controller_property_manager import ControllerPropertyManager
from ds_capability.intent.controller_intent import ControllerIntentModel
from ds_capability.components.discovery import DataDiscovery


# noinspection PyArgumentList
class Controller(AbstractComponent):
    """Controller Class for the management and overview of task components"""

    REPORT_USE_CASE = 'use_case'
    URI_PM_REPO = None

    def __init__(self, property_manager: ControllerPropertyManager, intent_model: ControllerIntentModel,
                 default_save=None, reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None,
                 align_connectors: bool=None):
        """ Encapsulation class for the components set of classes

        :param property_manager: The contract property manager instance for this component
        :param intent_model: the model codebase containing the parameterizable intent
        :param default_save: The default behaviour of persisting the contracts:
                    if False: The connector contracts are kept in memory (useful for restricted file systems)
        :param reset_templates: (optional) reset connector templates from environ variables (see `report_environ()`)
        :param template_path: (optional) a template path to use if the environment variable does not exist
        :param template_module: (optional) a template module to use if the environment variable does not exist
        :param template_source_handler: (optional) a template source handler to use if no environment variable
        :param template_persist_handler: (optional) a template persist handler to use if no environment variable
        :param align_connectors: (optional) resets aligned connectors to the template
        """
        super().__init__(property_manager=property_manager, intent_model=intent_model, default_save=default_save,
                         reset_templates=reset_templates, template_path=template_path, template_module=template_module,
                         template_source_handler=template_source_handler,
                         template_persist_handler=template_persist_handler, align_connectors=align_connectors)

    @classmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, creator: str, uri_pm_repo: str=None, pm_file_type: str=None,
                 pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None, default_save=None,
                 reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, has_contract: bool=None):
        """ Class Factory Method to instantiates the component's application. The Factory Method handles the
        instantiation of the Properties Manager, the Intent Model and the persistence of the uploaded properties.
        See class inline _docs for an example method

         :param task_name: The reference name that uniquely identifies a task or subset of the property manager
         :param uri_pm_path: A URI that identifies the resource path for the property manager.
         :param creator: A user name for this task activity.
         :param uri_pm_repo: (optional) A repository URI to initially load the property manager but not save to.
         :param pm_file_type: (optional) defines a specific file type for the property manager
         :param pm_module: (optional) the module or package name where the handler can be found
         :param pm_handler: (optional) the handler for retrieving the resource
         :param pm_kwargs: (optional) a dictionary of kwargs to pass to the property manager
         :param default_save: (optional) if the configuration should be persisted. default to 'True'
         :param reset_templates: (optional) reset connector templates from environ variables. Default True
                                (see `report_environ()`)
         :param template_path: (optional) a template path to use if the environment variable does not exist
         :param template_module: (optional) a template module to use if the environment variable does not exist
         :param template_source_handler: (optional) a template source handler to use if no environment variable
         :param template_persist_handler: (optional) a template persist handler to use if no environment variable
         :param align_connectors: (optional) resets aligned connectors to the template. default Default True
         :param default_save_intent: (optional) The default action for saving intent in the property manager
         :param default_intent_level: (optional) the default level intent should be saved at
         :param order_next_available: (optional) if the default behaviour for the order should be next available order
         :param default_replace_intent: (optional) the default replace existing intent behaviour
         :param has_contract: (optional) indicates the instance should have a property manager domain contract
         :return: the initialised class instance
         """
        # save the controllers uri_pm_repo path
        if isinstance(uri_pm_repo, str):
            cls.URI_PM_REPO = uri_pm_repo
        pm_file_type = pm_file_type if isinstance(pm_file_type, str) else 'json'
        pm_module = pm_module if isinstance(pm_module, str) else cls.DEFAULT_MODULE
        pm_handler = pm_handler if isinstance(pm_handler, str) else cls.DEFAULT_PERSIST_HANDLER
        _pm = ControllerPropertyManager(task_name=task_name, creator=creator)
        _intent_model = ControllerIntentModel(property_manager=_pm, default_save_intent=default_save_intent,
                                              default_intent_level=default_intent_level,
                                              order_next_available=order_next_available,
                                              default_replace_intent=default_replace_intent)
        super()._init_properties(property_manager=_pm, uri_pm_path=uri_pm_path, default_save=default_save,
                                 uri_pm_repo=uri_pm_repo, pm_file_type=pm_file_type, pm_module=pm_module,
                                 pm_handler=pm_handler, pm_kwargs=pm_kwargs, has_contract=has_contract)
        return cls(property_manager=_pm, intent_model=_intent_model, default_save=default_save,
                   reset_templates=reset_templates, template_path=template_path, template_module=template_module,
                   template_source_handler=template_source_handler, template_persist_handler=template_persist_handler,
                   align_connectors=align_connectors)

    @classmethod
    def from_env(cls, task_name: str=None, default_save=None, reset_templates: bool=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, uri_pm_repo: str=None, has_contract: bool=None,
                 **kwargs):
        """ Class Factory Method that builds the connector handlers taking the property contract path from
        the os.environ['HADRON_PM_PATH'] or, if not found, uses the system default,
                    for Linux and IOS '/tmp/components/contracts
                    for Windows 'os.environ['AppData']\\components\\contracts'
        The following environment variables can be set:
        'HADRON_PM_PATH': the property contract path, if not found, uses the system default
        'HADRON_PM_REPO': the property contract should be initially loaded from a read only repo site such as github
        'HADRON_PM_TYPE': a file type for the property manager. If not found sets as 'json'
        'HADRON_PM_MODULE': a default module package, if not set uses component default
        'HADRON_PM_HANDLER': a default handler. if not set uses component default

        This method calls to the Factory Method 'from_uri(...)' returning the initialised class instance

         :param task_name: (optional) The reference name that uniquely identifies the ledger. Defaults to 'primary'
         :param default_save: (optional) if the configuration should be persisted
         :param reset_templates: (optional) reset connector templates from environ variables. Default True
                                (see `report_environ()`)
         :param align_connectors: (optional) resets aligned connectors to the template. default Default True
         :param default_save_intent: (optional) The default action for saving intent in the property manager
         :param default_intent_level: (optional) the default level intent should be saved at
         :param order_next_available: (optional) if the default behaviour for the order should be next available order
         :param default_replace_intent: (optional) the default replace existing intent behaviour
         :param uri_pm_repo: The read only repo link that points to the raw data path to the contracts repo directory
         :param has_contract: (optional) indicates the instance should have a property manager domain contract
         :param kwargs: to pass to the property ConnectorContract as its kwargs
         :return: the initialised class instance
         """
        task_name = task_name if isinstance(task_name, str) else 'master'
        return super().from_env(task_name=task_name, default_save=default_save, reset_templates=reset_templates,
                                align_connectors=align_connectors, default_save_intent=default_save_intent,
                                default_intent_level=default_intent_level, order_next_available=order_next_available,
                                default_replace_intent=default_replace_intent, uri_pm_repo=uri_pm_repo,
                                has_contract=has_contract, **kwargs)

    @classmethod
    def scratch_pad(cls) -> ControllerIntentModel:
        """ A class method to use the Components intent methods as a scratch pad"""
        return super().scratch_pad()

    @property
    def intent_model(self) -> ControllerIntentModel:
        """The intent model instance"""
        return self._intent_model

    @property
    def register(self) -> ControllerIntentModel:
        """The intent model instance"""
        return self._intent_model

    @property
    def pm(self) -> ControllerPropertyManager:
        """The properties manager instance"""
        return self._component_pm

    def remove_all_tasks(self, save: bool=None):
        """removes all tasks"""
        for level in self.pm.get_intent():
            self.pm.remove_intent(level=level)
        self.pm_persist(save)

    def set_use_case(self, title: str=None, domain: str=None, overview: str=None, scope: str=None,
                     situation: str=None, opportunity: str=None, actions: str=None, project_name: str=None,
                     project_lead: str=None, project_contact: str=None, stakeholder_domain: str=None,
                     stakeholder_group: str=None, stakeholder_lead: str=None, stakeholder_contact: str=None,
                     save: bool=None):
        """ sets the use_case values. Only sets those passed

        :param title: (optional) the title of the use_case
        :param domain: (optional) the domain it sits within
        :param overview: (optional) a overview of the use case
        :param scope: (optional) the scope of responsibility
        :param situation: (optional) The inferred 'Why', 'What' or 'How' and predicted 'therefore can we'
        :param opportunity: (optional) The opportunity of the situation
        :param actions: (optional) the actions to fulfil the opportunity
        :param project_name: (optional) the name of the project this use case is for
        :param project_lead: (optional) the person who is project lead
        :param project_contact: (optional) the contact information for the project lead
        :param stakeholder_domain: (optional) the domain of the stakeholders
        :param stakeholder_group: (optional) the stakeholder group name
        :param stakeholder_lead: (optional) the stakeholder lead
        :param stakeholder_contact: (optional) contact information for the stakeholder lead
        :param save: (optional) if True, save to file. Default is True
        """
        self.pm.set_use_case(title=title, domain=domain, overview=overview, scope=scope, situation=situation,
                             opportunity=opportunity, actions=actions, project_name=project_name,
                             project_lead=project_lead, project_contact=project_contact,
                             stakeholder_domain=stakeholder_domain, stakeholder_group=stakeholder_group,
                             stakeholder_lead=stakeholder_lead, stakeholder_contact=stakeholder_contact)
        self.pm_persist(save=save)

    def load_source_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                              **kwargs) -> pa.Table:
        """returns the contracted source data as a DataFrame

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_SOURCE, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    def load_canonical(self, connector_name: str, reset_changed: bool=None, has_changed: bool=None,
                       return_empty: bool=None, **kwargs) -> pa.Table:
        """returns the canonical of the referenced connector

        :param connector_name: the name or label to identify and reference the connector
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        canonical = super().load_canonical(connector_name=connector_name, reset_changed=reset_changed,
                                           has_changed=has_changed, return_empty=return_empty, **kwargs)
        return canonical

    def load_persist_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                               **kwargs) -> pa.Table:
        """loads the clean pandas.DataFrame from the clean folder for this contract

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_PERSIST, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    @staticmethod
    def quality_report(canonical: pa.Table, nulls_threshold: float = None,
                       dom_threshold: float = None,
                       cat_threshold: int = None, stylise: bool = None):
        """ Analyses a dataset, passed as a DataFrame and returns a quality summary

        :param canonical: The table to view.
        :param cat_threshold: (optional) The threshold for the max number of unique categories. Default is 60
        :param dom_threshold: (optional) The threshold limit of a dominant value. Default 0.98
        :param nulls_threshold: (optional) The threshold limit of a nulls value. Default 0.9
        :param stylise: (optional) if the output is stylised
        """
        stylise = stylise if isinstance(stylise, bool) else True
        return DataDiscovery.data_quality(canonical=canonical, nulls_threshold=nulls_threshold,
                                          dom_threshold=dom_threshold, cat_threshold=cat_threshold,
                                          stylise=stylise)

    @staticmethod
    def canonical_report(canonical: pa.Table, headers: [str, list] = None,
                         regex: [str, list] = None, d_types: list = None,
                         drop: bool = None, stylise: bool = None, display_width: int = None):
        """The Canonical Report is a data dictionary of the canonical providing a reference view of the dataset's
        attribute properties

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param display_width: (optional) the width of the observational display
        """
        stylise = stylise if isinstance(stylise, bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types,
                                     drop=drop)
        return DataDiscovery.data_dictionary(canonical=tbl, stylise=stylise,
                                             display_width=display_width)

    @staticmethod
    def schema_report(canonical: pa.Table, headers: [str, list] = None, regex: [str, list] = None,
                      d_types: list = None,
                      drop: bool = None, stylise: bool = True, table_cast: bool = None):
        """ presents the current canonical schema

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param table_cast: (optional) if the column should try to be cast to its type
        """
        stylise = stylise if isinstance(stylise, bool) else True
        table_cast = table_cast if isinstance(table_cast, bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types,
                                     drop=drop)
        return DataDiscovery.data_schema(canonical=tbl, table_cast=table_cast, stylise=stylise)

    @staticmethod
    def table_report(canonical: pa.Table, head: int=None):
        """ Creates a report from a pyarrow table in a tabular form

        :param canonical: the table to view
        :param head: The number of rows to show. Default to 5
        """
        head = head if isinstance(head, int) else 5
        return Commons.table_report(canonical, head=head)

    def reset_use_case(self, save: bool=None):
        """resets the use_case back to its default values"""
        self.pm.reset_use_case()
        self.pm_persist(save)

    def report_use_case(self, stylise: bool=None):
        """ a report on the use_case set as part of the domain contract"""
        stylise = stylise if isinstance(stylise, bool) else True
        report = self.pm.report_use_case()
        report = pd.DataFrame(report, index=['values'])
        if stylise:
            report = report.transpose().reset_index()
            report.columns = ['use_case', 'values']
            return Commons.report(report, index_header='use_case')
        return pa.Table.from_pandas(report)

    def report_tasks(self, stylise: bool=True):
        """ generates a report for all the current component task"""
        report = pd.DataFrame.from_dict(data=self.pm.report_intent())
        intent_replace = {'feature_select': 'FeatureSelect', 'feature_engineer': 'FeatureEngineer',
                          'feature_transform': 'FeatureTransform', 'feature_build': 'FeatureBuild',
                          'automl': 'AutoML'}
        report['component'] = report.intent.replace(to_replace=intent_replace)
        report = report.loc[:, ['level', 'order', 'component', 'parameters', 'creator']]
        if stylise:
            return Commons.report(report, index_header='level')
        return pa.Table.from_pandas(report)

    def report_run_book(self, stylise: bool=True):
        """ generates a report on all the intent actions"""
        report = pd.DataFrame(self.pm.report_run_book())
        report = report.explode(column='run_book', ignore_index=True)
        report = report.join(pd.json_normalize(report['run_book'])).drop(columns=['run_book']).fillna('')
        if stylise:
            return Commons.report(report, index_header='name')
        return pa.Table.from_pandas(report)

    def add_run_book(self, run_levels: [str, list], book_name: str=None, save: bool=None):
        """ sets a named run book, the run levels are a list of levels and the order they are run in

        :param run_levels: the name or list of levels to be run
        :param book_name: (optional) the name of the run_book. defaults to 'primary_run_book'
        :param save: (optional) override of the default save action set at initialisation.
       """
        book_name = book_name if isinstance(book_name, str) else self.pm.PRIMARY_RUN_BOOK
        for run_level in Commons.list_formatter(run_levels):
            self.add_run_book_level(run_level=run_level, book_name=book_name)
        return

    def add_run_book_level(self, run_level: str, book_name: str=None, source: str=None, persist: [str, list]=None,
                           save: bool=None):
        """ adds a single runlevel to the end of a run_book. If the name already exists it will be replaced. When
        designing the runbook this gives easier implementation of the data handling specific for this runbook

        :param run_level: the run_level to add.
        :param book_name: (optional) the name of the run_book. defaults to 'primary_run_book'
        :param source: (optional) the intent level source
        :param persist: (optional) the intent level persist
        :param save: (optional) override of the default save action set at initialisation.
       """
        book_name = book_name if isinstance(book_name, str) else self.pm.PRIMARY_RUN_BOOK
        run_level = Commons.param2dict(run_level=run_level, source=source, persist=persist)
        if self.pm.has_run_book(book_name=book_name):
            run_levels = self.pm.get_run_book(book_name=book_name)
            for i in range(len(run_levels)):
                if isinstance(run_levels[i], str) and run_levels[i] == run_level:
                    run_levels[i].remove(run_level)
                elif isinstance(run_levels[i], dict) and run_levels[i].get('run_level') == run_level.get('run_level'):
                    del run_levels[i]
            run_levels.append(run_level)
        else:
            run_levels = [run_level]
        self.pm.set_run_book(book_name=book_name, run_levels=run_levels)
        self.pm_persist(save)


    def report_intent(self, levels: [str, int, list]=None, stylise: bool = True):
        """ generates a report on all the intent

        :param levels: (optional) a filter on the levels. passing a single value will report a single parameterised view
        :param stylise: (optional) returns a stylised dataframe with formatting
        :return: pd.Dataframe
        """
        if isinstance(levels, (int, str)):
            report = pd.DataFrame.from_dict(data=self.pm.report_intent_params(level=levels))
            if stylise:
                return Commons.report(report, index_header='order')
        else:
            report = pd.DataFrame.from_dict(data=self.pm.report_intent(levels=levels))
            if stylise:
                return Commons.report(report, index_header='level')
        return pa.Table.from_pandas(report)

    def report_notes(self, catalog: [str, list]=None, labels: [str, list]=None, regex: [str, list]=None,
                     re_ignore_case: bool=False, stylise: bool=True, drop_dates: bool=False):
        """ generates a report on the notes

        :param catalog: (optional) the catalog to filter on
        :param labels: (optional) s label or list of labels to filter on
        :param regex: (optional) a regular expression on the notes
        :param re_ignore_case: (optional) if the regular expression should be case sensitive
        :param stylise: (optional) returns a stylised dataframe with formatting
        :param drop_dates: (optional) excludes the 'date' column from the report
        :return: pd.Dataframe
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        drop_dates = False if not isinstance(drop_dates, bool) else drop_dates
        report = self.pm.report_notes(catalog=catalog, labels=labels, regex=regex, re_ignore_case=re_ignore_case,
                                      drop_dates=drop_dates)
        report = pd.DataFrame.from_dict(data=report)
        if stylise:
            return Commons.report(report, index_header='section', large_font='label')
        return pa.Table.from_pandas(report)

    def run_controller(self, run_book: [str, list, dict]=None, repeat: int=None, sleep: int=None, run_time: int=None,
                       source_check_uri: str=None, run_cycle_report: str=None):
        """ Runs the components pipeline based on the runbook instructions. The run_book is a pre-registered
        Controller run_book names to execute. If no run book is given, default values are substituted, finally taking
        the intent list if all else fails.

        The run_cycle_report automatically generates the connector contract with the name 'run_cycle_report'. To reload
        the report for observation use the controller method 'load_canonical(...) passing the name 'run_cycle_report'.

        :param run_book: (optional) a run_book reference, a list of task names (intent levels)
        :param repeat: (optional) the number of times this intent should be repeated. None or -1 -> never, 0 -> forever
        :param sleep: (optional) number of seconds to sleep before repeating
        :param run_time: (optional) number of seconds to run the controller using repeat and sleep cycles time is up
        :param source_check_uri: (optional) The source uri to check for change since last controller instance cycle
        :param run_cycle_report: (optional) a full name for the run cycle report
        """
        if isinstance(run_cycle_report, str):
            self.add_connector_persist(connector_name='run_cycle_report', uri_file=run_cycle_report)
        df_report = pd.DataFrame(columns=['time', 'text'])
        if not self.pm.has_intent():
            return
        # tidy run_book
        run_book = Commons.list_formatter(run_book)
        if not run_book:
            if self.pm.has_run_book(self.pm.PRIMARY_RUN_BOOK):
                run_book = self.pm.get_run_book(self.pm.PRIMARY_RUN_BOOK)
            elif self.pm.has_intent(self.pm.DEFAULT_INTENT_LEVEL):
                self.add_run_book_level(run_level=self.pm.DEFAULT_INTENT_LEVEL, book_name=self.pm.PRIMARY_RUN_BOOK)
                run_book = self.pm.get_run_book(self.pm.PRIMARY_RUN_BOOK)
            else:
                for level in self.pm.get_intent().keys():
                    self.add_run_book_level(run_level=level, book_name=self.pm.PRIMARY_RUN_BOOK)
                run_book = self.pm.get_run_book(self.pm.PRIMARY_RUN_BOOK)
        run_dict = []
        for book in run_book:
            if isinstance(book, str):
                if self.pm.has_run_book(book):
                    for item in self.pm.get_run_book(book):
                        run_dict.append(item)
                else:
                    run_dict.append({'run_level': f'{book}'})
            else:
                run_dict.append(book)
        handler = None
        if isinstance(source_check_uri, str):
            self.add_connector_uri(connector_name='source_check_uri', uri=source_check_uri)
            handler = self.pm.get_connector_handler(connector_name='source_check_uri')
        run_time = run_time if isinstance(run_time, int) else 0
        if run_time > 0 and not isinstance(sleep, int):
            sleep = 1
        end_time = datetime.datetime.now() + datetime.timedelta(seconds=run_time)
        repeat = repeat if isinstance(repeat, int) and repeat > 0 else 1
        run_count = 0
        while True: # run_time always runs once
            if isinstance(run_cycle_report, str):
                df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f'starting pipeline -{run_count}-']
            for count in range(repeat):
                if isinstance(run_cycle_report, str):
                    df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f'starting cycle -{count}-']
                if handler and handler.exists():
                    if handler.has_changed():
                        handler.reset_changed(False)
                    else:
                        if isinstance(run_cycle_report, str):
                            df_report.loc[len(df_report.index)] = [datetime.datetime.now(), 'Source has not changed']
                        if isinstance(sleep, int) and count < repeat - 1:
                            time.sleep(sleep)
                        continue
                for book in run_dict:
                    run_level = book.get('run_level')
                    source = book.get('source', self.CONNECTOR_SOURCE)
                    persist = book.get('persist', self.CONNECTOR_PERSIST)
                    if isinstance(run_cycle_report, str):
                        df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f"running: '{run_level}'"]
                    # run level
                    shape = self.intent_model.run_intent_pipeline(run_level=run_level, source=source, persist=persist,
                                                                  controller_repo=self.URI_PM_REPO)
                    if isinstance(run_cycle_report, str):
                        df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f'outcome shape: {shape}']
                if isinstance(run_cycle_report, str):
                    df_report.loc[len(df_report.index)] = [datetime.datetime.now(), 'cycle complete']
                if isinstance(sleep, int) and count < repeat-1:
                    if isinstance(run_cycle_report, str):
                        df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f'sleep for {sleep} seconds']
                    time.sleep(sleep)
            if isinstance(run_cycle_report, str):
                run_count += 1
            if end_time < datetime.datetime.now():
                break
            else:
                if isinstance(run_cycle_report, str):
                    df_report.loc[len(df_report.index)] = [datetime.datetime.now(), f'sleep for {sleep} seconds']
                time.sleep(sleep)
        if isinstance(run_cycle_report, str):
            df_report.loc[len(df_report.index)] = [datetime.datetime.now(), 'pipeline complete']
            _ = pa.Table.from_pandas(df_report, preserve_index=False)
            self.save_canonical(connector_name='run_cycle_report', canonical=_)
        return

    @staticmethod
    def runbook_script(task: str, source: str=None, persist: str=None, drop: bool=None) -> dict:
        """ a utility method to help build feature conditions by aligning method parameters with dictionary format.
        This also sets default, in-memory and connector contracts specific for a task.
        if not specified, source or persist will use the default values. '${<connector>}' will use the environment
        variable is a connector contract of that name that must exist in the task connectors.

        :param task: the task name (intent level) name this runbook is applied too or a number if synthetic generation.
        :param source: (optional) a task name indicating where the source of this task will come from.
        :param persist: (optional) a task name indicating where the persist of this task will go to.
        :param drop: (optional) if true indicates in-memory canonical should be dropped once read.
        :return: dictionary of the parameters
        """
        source = source if isinstance(source, str) else Controller.CONNECTOR_SOURCE
        source = ConnectorContract.parse_environ(source) if source.startswith('${') and source.endswith('}') else source
        persist = persist if isinstance(persist, str) else Controller.CONNECTOR_PERSIST
        persist = ConnectorContract.parse_environ(persist) if persist.startswith('${') and persist.endswith('}') else persist
        drop = drop if isinstance(drop, bool) else True
        return Commons.param2dict(**locals())




from ds_core.properties.abstract_properties import AbstractPropertyManager
from ds_capability.components.commons import Commons

__author__ = 'Darryl Oatridge'


class AutoMLPropertyManager(AbstractPropertyManager):

    def __init__(self, task_name: str, creator: str, root_keys: list=None, knowledge_keys: list=None):
        root_keys = root_keys if isinstance(root_keys, list) else []
        knowledge_keys = knowledge_keys if isinstance(knowledge_keys, list) else []
        # set additional keys
        r_extended = Commons.list_unique(root_keys + [])
        k_extended = Commons.list_unique(knowledge_keys + ['features', 'observations', 'models'])
        super().__init__(task_name=task_name, root_keys=r_extended, knowledge_keys=k_extended, creator=creator)

    @staticmethod
    def list_formatter(value) -> list:
        """override of the list_formatter to include Pandas types"""
        return Commons.list_formatter(value=value)



import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_core.properties.property_manager import PropertyManager
from ds_capability import *
from ds_capability.components.commons import Commons

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class TemplateTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_run_data_tyoe_runbook(self):
        os.environ['HADRON_PROFILING_SOURCE_URI'] = 'working/source/hadron_synth_other.pq'
        os.environ['HADRON_DATA_QUALITY_URI'] = 'working/data/quality.parquet'
        os.environ['HADRON_DATA_DICTIONARY_URI'] = 'working/data/dictionary.parquet'
        os.environ['HADRON_DATA_SCHEMA_URI'] = 'working/data/schema.parquet'

        from ds_capability import FeatureBuild
        from ds_capability.components.commons import Commons

        # feature builder
        fb = FeatureBuild.from_env('data_profiling', has_contract=False)
        tbl = fb.set_source_uri('${HADRON_PROFILING_SOURCE_URI}').load_source_canonical()
        _ = fb.set_persist_uri('event://profiling')
        _ = fb.add_connector_uri('quality', '${HADRON_DATA_QUALITY_URI}')
        _ = fb.add_connector_uri('dictionary', '${HADRON_DATA_DICTIONARY_URI}')
        _ = fb.add_connector_uri('schema', '${HADRON_DATA_SCHEMA_URI}')
        tbl = fb.tools.build_profiling(tbl, profiling='quality', connector_name='quality', intent_order=0)
        tbl = fb.tools.build_profiling(tbl, profiling='dictionary', connector_name='dictionary', intent_order=1)
        tbl = fb.tools.build_profiling(tbl, profiling='schema', connector_name='schema', intent_order=2)
        fb.run_component_pipeline()
        print(fb.pm.get_all())


        # c = Controller.from_env()
        # print(pm_view('controller', 'master'))
        # c.run_controller()

    def test_run_feature_select(self):
        fe = FeatureEngineer.from_memory()
        fe.set_persist('synthetic_data.parquet')
        tbl = fe.tools.get_synthetic_data_types(10)
        fe.save_persist_canonical(tbl)
        # test
        fs = FeatureSelect.from_env('fs_component', has_contract=False)
        fs.set_source(fe.get_persist_uri())
        fs.set_persist()
        print(fs.report_connectors().to_string())

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def tprint(t: pa.table, headers: [str, list] = None, d_type: [str, list] = None, regex: [str, list] = None):
    _ = Commons.filter_columns(t.slice(0, 10), headers=headers, d_types=d_type, regex=regex)
    print(Commons.table_report(_).to_string())


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime
from pprint import pprint

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_core.handlers.event_handlers import EventPersistHandler

from ds_capability.intent.feature_engineer_intent import FeatureEngineerIntent
from ds_core.properties.property_manager import PropertyManager

from ds_capability import FeatureEngineer

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureEngineerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_run_intent_pipeline(self):
        fb = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fb.tools
        _ = tools.get_synthetic_data_types(size=10, extend=True, intent_level='simulator')
        _ = tools.correlate_number(_, header='num', intent_level='data_quality', intent_order=0)
        # pprint(pm_view('feature_build', 'test', 'intent'))
        fb.run_component_pipeline(intent_levels=['simulator', 'data_quality'])
        result = fb.load_persist_canonical()
        self.assertEqual((10, 20), result.shape)

    def test_run_intent_pipeline_event_manager(self):
        fb = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fb.tools
        fb.set_persist_uri('event://task')
        _ = tools.get_synthetic_data_types(size=10, extend=True, intent_level='simulator')
        _ = tools.correlate_number(_, header='num', intent_level='data_quality', intent_order=0)
        # pprint(pm_view('feature_build', 'test', 'intent'))
        fb.run_component_pipeline(intent_levels=['simulator', 'data_quality'])
        result = fb.load_persist_canonical()
        self.assertEqual((10,20), result.shape)
        h = fb.pm.get_connector_handler(fb.CONNECTOR_PERSIST)
        self.assertIsInstance(h,EventPersistHandler)

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def get_table():
    num = pa.array([1.0, None, 5.0, -0.46421, 3.5, 7.233, -2], pa.float64())
    val = pa.array([1, 2, 3, 4, 5, 6, 7], pa.int64())
    date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, None, "2023-01-02 05:23:50", None, None],
                       format='%Y-%m-%d %H:%M:%S', unit='ns')
    text = pa.array(["Blue", "Green", None, 'Red', 'Orange', 'Yellow', 'Pink'], pa.string())
    binary = pa.array([True, True, None, False, False, True, False], pa.bool_())
    cat = pa.array([None, 'M', 'F', 'M', 'F', 'M', 'M'], pa.string()).dictionary_encode()
    return pa.table([num, val, date, text, binary, cat], names=['num', 'int', 'date', 'text', 'bool', 'cat'])


def pm_view(capability: str, task: str, section: str = None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability, {}).get(task, {})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result


if __name__ == '__main__':
    unittest.main()



import time
import numpy as np
import pandas as pd
from ds_core.handlers.abstract_handlers import ConnectorContract

__author__ = 'Darryl Oatridge'


class CommonsIntentModel(object):

    @classmethod
    def __dir__(cls):
        """returns the list of available methods associated with the parameterized intent"""
        rtn_list = []
        for m in dir(cls):
            if not m.startswith('_'):
                rtn_list.append(m)
        return rtn_list

    """
        PRIVATE METHODS SECTION
    """

    @staticmethod
    def _seed(seed: int=None, increment: bool=False):
        if not isinstance(seed, int):
            return int(time.time() * np.random.default_rng().random())
        if increment:
            seed += 1
            if seed > 2 ** 31:
                seed = int(time.time() * np.random.default_rng(seed=seed-1).random())
        return seed

    def _set_quantity(self, selection, quantity, seed=None):
        """Returns the quantity percent of good values in selection with the rest fill"""
        quantity = self._quantity(quantity)
        if quantity == 1:
            return selection
        if quantity == 0:
            return [np.nan] * len(selection)
        seed = self._seed(seed=seed)
        quantity = 1 - quantity
        generator = np.random.default_rng(seed)
        length = len(selection)
        size = int(length * quantity)
        nulls_idx = generator.choice(length, size=size, replace=False)
        result = pd.Series(selection)
        result.iloc[nulls_idx] = np.nan
        return result.to_list()

    @staticmethod
    def _quantity(quantity: [float, int]) -> float:
        """normalises quantity to a percentate float between 0 and 1.0"""
        if not isinstance(quantity, (int, float)) or not 0 <= quantity <= 100:
            return 1.0
        if quantity > 1:
            return round(quantity / 100, 2)
        return float(quantity)


    @staticmethod
    def _extract_value(value: [str, int, float]):
        if isinstance(value, str):
            if value.startswith('${') and value.endswith('}'):
                value = ConnectorContract.parse_environ(value)
                if value.isnumeric():
                    return int(value)
                elif value.replace('.', '', 1).isnumeric():
                    return float(value)
                else:
                    return str(value)
            else:
                return str(value)
        return value

    """
        UTILITY METHODS SECTION
    """

    @staticmethod
    def _freq_dist_size(relative_freq: list, size: int, dist_length: int=None, dist_on: str=None, seed: int=None):
        """ utility method taking a list of relative frequencies and based on size returns the size distribution
        of element based on the frequency. The distribution is based upon binomial distributions.

        :param relative_freq: a list of int or float values representing a relative distribution frequency
        :param size: the size of the values to be distributed
        :param dist_length: (optional) the expected length of the element's in relative_freq
        :param dist_on: (optional) if element length differs. distribute on 'left', 'right' or 'center'. Default 'right'
        :param seed: (optional) a seed value for the random function: default to None
        :return: an integer list of the distribution that sum to the size
        """
        if not isinstance(relative_freq, list) or not all(isinstance(x, (int, float)) for x in relative_freq):
            raise ValueError("The weighted pattern must be an list of numbers")
        dist_length = dist_length if isinstance(dist_length, int) else len(relative_freq)
        dist_on = dist_on if dist_on in ['right', 'left'] else 'both'
        seed = seed if isinstance(seed, int) else int(time.time() * np.random.random())
        # sort the width
        if len(relative_freq) > dist_length:
            a = np.array(relative_freq)
            trim = dist_length - a.size
            if dist_on.startswith('right'):
                relative_freq = a[:trim]
            elif dist_on.startswith('left'):
                relative_freq = a[abs(trim):]
            else:
                l_size = int(trim/2)
                r_size = a.size + l_size - dist_length
                relative_freq = a[r_size:l_size].tolist()
        if len(relative_freq) < dist_length:
            a = np.array(relative_freq)
            rvalue = a[-1:]
            lvalue = a[:1]
            if dist_on.startswith('left'):
                r_dist = []
                l_dist = np.tile(lvalue, dist_length - a.size)
            elif dist_on.startswith('right'):
                r_dist = np.tile(rvalue, dist_length - a.size)
                l_dist = []
            else:
                l_size = int((dist_length - a.size) / 2)
                r_dist = np.tile(rvalue, l_size)
                l_dist = np.tile(lvalue, dist_length - l_size - a.size)
            relative_freq = np.hstack([l_dist, a, r_dist]).tolist()
        # turn it to percentage
        if sum(relative_freq) != 1:
            relative_freq = np.round(relative_freq / np.sum(relative_freq), 5)
        generator = np.random.default_rng(seed=seed)
        result = list(generator.binomial(n=size, p=relative_freq, size=len(relative_freq)))
        diff = size - sum(result)
        adjust = [0] * len(relative_freq)
        # There is a possibility the required size is not fulfilled, therefore add or remove elements based on freq
        if diff != 0:
            unit = diff / sum(relative_freq)
            for idx in range(len(relative_freq)):
                adjust[idx] = int(round(relative_freq[idx] * unit, 0))
        result = [a + b for (a, b) in zip(result, adjust)]
        # rounding can still make us out by 1
        if sum(result) != size:
            gap = sum(result) - size
            result[result.index(max(result))] -= gap
        return result[:size]



import io
import os
import threading
from typing import Optional
import pyarrow as pa
import pyarrow.csv as csv
import pyarrow.parquet as pq
from .cortex_helpers import load_token, load_api_endpoint
from ds_core.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract, AbstractPersistHandler, HandlerFactory


class McSourceHandler(AbstractSourceHandler):
    """ A Managed Content Source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ Initialise the handler passing the source_contract dictionary """
        super().__init__(connector_contract)
        self.cortex = HandlerFactory.get_module('cortex')
        self.token = self._load_token()
        self.api_endpoint = self._load_api_endpoint()
        self.project = self._load_project_name()
        self.cortex_client = self.cortex.Cortex.client(project=self.project)
        # _ = self.cortex_client.content.exists
        self._etag = 0
        self._changed_flag = True

    def mc_key(self, connector_contract: Optional[ConnectorContract]=None):
        _cc = connector_contract if connector_contract is not None else self.connector_contract
        schema, bucket, path = _cc.parse_address_elements(_cc.uri)
        if not path:
            return bucket
        return os.path.join(bucket, path.strip('/'))

    def _load_token(self):
        return load_token(token=self.connector_contract.kwargs.get("token", os.environ["TOKEN"]))

    def _load_api_endpoint(self):
        return load_api_endpoint(endpoint=self.connector_contract.kwargs.get("api_endpoint", os.environ["API_ENDPOINT"]))
    
    def _load_project_name(self):
        return self.connector_contract.kwargs.get("project", os.environ["PROJECT"])

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ["csv", "parquet"]

    def _download_key_from_mc(self, key):
        return self.cortex_client.content.download(key, retries=2)

    def _load_df_from_csv_in_mc(self, mc_key: str, **options) -> pa.Table:
        content = self._download_key_from_mc(mc_key).read()
        return csv.read_csv(io.BytesIO(content), **options)

    def _load_df_from_parquet_in_mc(self, mc_key: str, **options) -> pa.Table:
        content = self._download_key_from_mc(mc_key).read()
        return pq.read_table(io.BytesIO(content), **options)

    def load_canonical(self) -> pa.Table:
        """ returns the canonical dataset based on the connector contract. This method utilises the pandas
        'pd.read_' methods and directly passes the kwargs to these methods.
        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Managed Content Connector Contract has not been set")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        load_params = _cc.kwargs
        load_params.update(_cc.query)  # Update kwargs with those in the uri query
        load_params.pop('token', None)
        load_params.pop('api_endpoint', None)
        load_params.pop('project', None)
        _, _, _ext = _cc.address.rpartition('.')
        file_type = load_params.get('file_type', _ext if len(_ext) > 0 else 'csv')
        if file_type.lower() not in self.supported_types():
            raise ValueError("The file type {} is not recognised. "
                             "Set file_type parameter to a recognised source type".format(file_type))

        # session
        if _cc.schema not in ['mc']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        with threading.Lock():
            if file_type.lower() in ['csv', 'gz']:
                rtn_data = self._load_df_from_csv_in_mc(mc_key=self.mc_key(), **load_params)
            elif file_type.lower() in ['parquet']:
                rtn_data = self._load_df_from_parquet_in_mc(mc_key=self.mc_key(),
                **load_params)
            else:
                raise LookupError('The source format {} is not currently supported'.format(file_type))
        self.reset_changed()
        return rtn_data

    def exists(self) -> bool:
        """ returns True if the file in mc exists """
        _cc = self.connector_contract
        mc_key = self.mc_key()
        return self.cortex_client.content.exists(key=mc_key)

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed
    
    def has_changed(self) -> bool:
        """ 
            returns if the file has been modified
            uses etag
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Managed Content Connector Contract has not been set")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        load_params = _cc.kwargs
        load_params.update(_cc.query)  # Update kwargs with those in the uri query
        load_params.pop('token', None)
        load_params.pop('api_endpoint', None)
        load_params.pop('project', None)
        res = self._download_key_from_mc(self.mc_key())  
        _etag = res.headers['etag']
        if _etag != self._etag:
            self._changed_flag = True
            self._etag = _etag
        else:
            self._changed_flag = False
        return self._changed_flag

class McPersistHandler(McSourceHandler, AbstractPersistHandler):
    # A Managed Content persist handler

    def _persist_df_as_csv(self, canonical: pa.Table, mc_key: str, **kwargs):
        file_name = os.path.basename(mc_key)
        byte_obj = io.BytesIO()
        csv.write_csv(canonical, byte_obj)
        res = self.cortex_client.content.upload(key=mc_key, stream_name=file_name, stream=byte_obj, content_type="application/octet-stream", retries=2)
        return res
    
    def _persist_df_as_parquet(self, canonical: pa.Table, mc_key: str, **kwargs):
        file_name = os.path.basename(mc_key)
        pq.write_table(canonical, file_name)
        with open(file_name, mode="rb") as f_obj:
            res = self.cortex_client.content.upload(key=mc_key, stream_name=file_name, stream=f_obj, content_type="application/octet-stream", retries=2)
        return res

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset
        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _uri = self.connector_contract.uri
        return self.backup_canonical(uri=_uri, canonical=canonical)

    def backup_canonical(self, canonical: pa.Table, uri: str, ignore_kwargs: bool = False) -> bool:
        """ creates a backup of the canonical to an alternative URI  """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        schema, bucket, path = _cc.parse_address_elements(uri=uri)
        _, _, _ext = path.rpartition('.')
        load_params = _cc.kwargs
        load_params.update(_cc.query)  # Update kwargs with those in the uri query
        load_params.pop('token', None)
        load_params.pop('api_endpoint', None)
        load_params.pop('project', None)
        mc_key = self.mc_key()
        file_type = load_params.get('file_type', _ext if len(_ext) > 0 else 'parquet')
        with threading.Lock():
            if file_type.lower() in ['csv']:
                self._persist_df_as_csv(canonical, mc_key=mc_key, **load_params)
            elif file_type.lower() in ['parquet']:
                self._persist_df_as_parquet(canonical=canonical, mc_key=mc_key)
            else:
                raise LookupError('The source format {} is not currently supported'.format(file_type))

        return True

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Managed Content Connector Contract has not been set")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        load_params = _cc.kwargs
        load_params.update(_cc.query)  # Update kwargs with those in the uri query
        load_params.pop('token', None)
        load_params.pop('api_endpoint', None)
        load_params.pop('project', None)
        self.cortex_client.content.delete(self.mc_key())
        if not self.exists():
            return True
        return False



data = ['You took too long', 'You are not easy to deal with', 'Payment Failure/Incorrect Payment',
        'You provided me with incorrect information', 'Unhappy with delay', 'Unsuitable advice',
        'You never answered my question', 'You did not understand my needs', 'I have been mis-sold',
        'My details are not accurate', 'You have asked for too much information', 'You were not helpful',
        'Payment not generated/received by customer', 'You did not keep me updated', 'Incorrect information given',
        'The performance of my product was poor', 'No reply to customer contact', 'Requested documentation not issued',
        'You did not explain the terms & conditions', 'Policy amendments not carried out',
        'You did not explain the next steps/process to me', 'I cannot understand your letter/comms',
        'Standard letter inappropriate', 'Customer payment processed incorrectly', 'All points not addressed',
        'Could not understand the agent', 'Issue with terms and conditions', 'Misleading information',
        'I can not use the customer portal', 'your customer portal is unhelpful']



import inspect
import pandas as pd
import pyarrow as pa
from ds_capability.components.discovery import DataDiscovery
from ds_capability.intent.common_intent import CommonsIntentModel
from ds_capability.intent.abstract_feature_build_intent import AbstractFeatureBuildIntentModel
from ds_capability.components.commons import Commons


class FeatureBuildIntent(AbstractFeatureBuildIntentModel, CommonsIntentModel):

    """This class is for feature builds intent actions which are bespoke to a certain used case
    but have broader reuse beyond this use case.
    """

    def build_difference(self, canonical: pa.Table, other: [str, pa.Table], on_key: [str, list], drop_zero_sum: bool=None,
                         summary_connector: bool=None, flagged_connector: str=None, detail_connector: str=None,
                         unmatched_connector: str=None, seed: int=None, save_intent: bool=None,
                         intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                         remove_duplicates: bool=None) -> pa.Table:
        """returns the difference between two canonicals, joined on a common and unique key.
        The ``on_key`` parameter can be a direct reference to the canonical column header or to an environment
        variable. If the environment variable is used ``on_key`` should be set to ``"${<<YOUR_ENVIRON>>}"`` where
        <<YOUR_ENVIRON>> is the environment variable name.

        If the ``flagged connector`` parameter is used, a report flagging mismatched left data with right data
        is produced for this connector where 1 indicate a difference and 0 they are the same. By default this method
        returns this report but if this parameter is set the original canonical returned. This allows a canonical
        pipeline to continue through the component while outputting the difference report.

        If the ``detail connector`` parameter is used, a detail report of the difference where the left and right
        values that differ are shown.

        If the ``unmatched connector`` parameter is used, the on_key's that don't match between left and right are
        reported

        :param canonical: a pa.Table as the reference table
        :param other: a direct pa.Table or reference to a connector.
        :param on_key: The name of the key that uniquely joins the canonical to others
        :param drop_zero_sum: (optional) drops rows and columns which has a total sum of zero differences
        :param summary_connector: (optional) a connector name where the summary report is sent
        :param flagged_connector: (optional) a connector name where the differences are flagged
        :param detail_connector: (optional) a connector name where the differences are shown
        :param unmatched_connector: (optional) a connector name where the unmatched keys are shown
        :param seed: (optional) this is a placeholder, here for compatibility across methods
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the column name that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        # intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intent action
        canonical = self._get_canonical(canonical)
        other = self._get_canonical(other)
        seed = seed if isinstance(seed, int) else self._seed()
        drop_zero_sum = drop_zero_sum if isinstance(drop_zero_sum, bool) else False
        flagged_connector = self._extract_value(flagged_connector)
        summary_connector = self._extract_value(summary_connector)
        detail_connector = self._extract_value(detail_connector)
        unmatched_connector = self._extract_value(unmatched_connector)
        on_key = Commons.list_formatter(self._extract_value(on_key))
        left_diff = Commons.list_diff(canonical.column_names, other.column_names, symmetric=False)
        right_diff = Commons.list_diff(other.column_names, canonical.column_names, symmetric=False)
        # drop columns
        tbl_canonical = canonical.drop_columns(left_diff)
        tbl_other = other.drop_columns(right_diff)
        # pandas
        df_canonical = tbl_canonical.to_pandas()
        df_other = tbl_other.to_pandas()
        # sort
        df_canonical.sort_values(on_key, inplace=True)
        df_other.sort_values(on_key, inplace=True)
        df_other = df_other.loc[:, df_canonical.columns.to_list()]

        # unmatched report
        if isinstance(unmatched_connector, str):
            if self._pm.has_connector(unmatched_connector):
                left_merge = pd.merge(df_canonical, df_other, on=on_key, how='left', suffixes=('', '_y'), indicator=True)
                left_merge = left_merge[left_merge['_merge'] == 'left_only']
                left_merge = left_merge[left_merge.columns[~left_merge.columns.str.endswith('_y')]]
                right_merge = pd.merge(df_canonical, df_other, on=on_key, how='right', suffixes=('_y', ''), indicator=True)
                right_merge = right_merge[right_merge['_merge'] == 'right_only']
                right_merge = right_merge[right_merge.columns[~right_merge.columns.str.endswith('_y')]]
                unmatched = pd.concat([left_merge, right_merge], axis=0, ignore_index=True)
                unmatched = unmatched.set_index(on_key, drop=True).reset_index(drop=True)
                unmatched.insert(0, 'found_in', unmatched.pop('_merge'))
                handler = self._pm.get_connector_handler(unmatched_connector)
                handler.persist_canonical(pa.Table.from_pandas(unmatched))
            else:
                raise ValueError(f"The connector name {unmatched_connector} has been given but no Connect Contract added")

        # remove non-matching rows
        df = pd.merge(df_canonical, df_other, on=on_key, how='inner', suffixes=('_x', '_y'))
        df_x = df.filter(regex='(_x$)', axis=1)
        df_y = df.filter(regex='(_y$)', axis=1)
        df_x.columns = df_x.columns.str.removesuffix('_x')
        df_y.columns = df_y.columns.str.removesuffix('_y')
        # flag the differences
        diff = df_x.ne(df_y).astype(int)
        if drop_zero_sum:
            diff = diff.loc[(diff != 0).any(axis=1),(diff != 0).any(axis=0)]
        # add back the keys
        for n in range(len(on_key)):
            diff.insert(n, on_key[n], df[on_key[n]].iloc[diff.index])

        # detailed report
        if isinstance(detail_connector, str):
            if self._pm.has_connector(detail_connector):
                diff_comp = df_x.astype(str).compare(df_y.astype(str)).fillna('-')
                for n in range(len(on_key)):
                    diff_comp.insert(n, on_key[n], df[on_key[n]].iloc[diff_comp.index])
                diff_comp.columns = ['_'.join(col) for col in diff_comp.columns.values]
                diff_comp.columns = diff_comp.columns.str.replace(r'_self$', '_x', regex=True)
                diff_comp.columns = diff_comp.columns.str.replace(r'_other$', '_y', regex=True)
                diff_comp.columns = diff_comp.columns.str.replace(r'_$', '', regex=True)
                diff_comp = diff_comp.sort_values(on_key)
                diff_comp = diff_comp.reset_index(drop=True)
                handler = self._pm.get_connector_handler(detail_connector)
                handler.persist_canonical(pa.Table.from_pandas(diff_comp))
            else:
                raise ValueError(f"The connector name {detail_connector} has been given but no Connect Contract added")

        # summary report
        if isinstance(summary_connector, str):
            if self._pm.has_connector(summary_connector):
                summary = diff.drop(on_key, axis=1).sum().reset_index()
                summary.columns = ['Attribute', 'Summary']
                summary = summary.sort_values(['Attribute'])
                indicator = pd.merge(df_canonical[on_key], df_other[on_key], on=on_key, how='outer', indicator=True)
                count = indicator['_merge'].value_counts().to_frame().reset_index().replace('both', 'matching')
                count.columns = ['Attribute', 'Summary']
                summary = pd.concat([count, summary], axis=0).reset_index(drop=True)
                handler = self._pm.get_connector_handler(summary_connector)
                handler.persist_canonical(pa.Table.from_pandas(summary))
            else:
                raise ValueError(f"The connector name {summary_connector} has been given but no Connect Contract added")

        # flagged report
        if isinstance(flagged_connector, str):
            if self._pm.has_connector(flagged_connector):
                diff = diff.sort_values(on_key)
                diff = diff.reset_index(drop=True)
                handler = self._pm.get_connector_handler(flagged_connector)
                handler.persist_canonical(pa.Table.from_pandas(diff))
                return canonical
            raise ValueError(f"The connector name {flagged_connector} has been given but no Connect Contract added")

        if drop_zero_sum:
            diff = diff.sort_values(on_key)
            diff = diff.reset_index(drop=True)

        return pa.Table.from_pandas(diff)

    def build_profiling(self, canonical: pa.Table, profiling: str, headers: [str, list]=None, d_types: [str, list]=None,
                        regex: [str, list]=None, drop: bool=None, connector_name: str=None, seed: int=None,
                        save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                        replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ Data profiling provides, analyzing, and creating useful summaries of data. The process yields a high-level
        overview which aids in the discovery of data quality issues, risks, and overall trends. It can be used to
        identify any errors, anomalies, or patterns that may exist within the data. There are three types of data
        profiling available 'dictionary', 'schema' or 'quality'

        :param canonical: a direct or generated pd.DataFrame. see context notes below
        :param profiling: The profiling name. Options are 'dictionary', 'schema' or 'quality'
        :param headers: (optional) a filter of headers from the 'other' dataset
        :param d_types: (optional) a filter on data type for the 'other' dataset. int, float, bool, object
        :param regex: (optional) a regular expression to search the headers. example '^((?!_amt).)*$)' excludes '_amt'
        :param drop: (optional) to drop or not drop the headers if specified
        :param connector_name: (optional) a connector name where the outcome is sent
        :param seed: (optional) this is a placeholder, here for compatibility across methods
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the column name that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        # intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intent action
        canonical = self._get_canonical(canonical)
        connector_name = self._extract_value(connector_name)
        if d_types is None:
            d_types = []
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        _seed = self._seed() if seed is None else seed
        if profiling == 'dictionary':
            result =  DataDiscovery.data_dictionary(canonical=canonical, table_cast=True, stylise=False)
        elif profiling == 'quality':
            result =  DataDiscovery.data_quality(canonical=canonical, stylise=False)
        elif profiling == 'schema':
            result = DataDiscovery.data_schema(canonical=canonical, stylise=False)
        else:
            raise ValueError(f"The report name '{profiling}' is not recognised. Use 'dictionary', 'schema' or 'quality'")
        if isinstance(connector_name, str):
            if self._pm.has_connector(connector_name):
                handler = self._pm.get_connector_handler(connector_name)
                handler.persist_canonical(result)
                return canonical
            raise ValueError(f"The connector name {connector_name} has been given but no Connect Contract added")
        return result



import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_capability.intent.feature_build_intent import FeatureBuildIntent
from ds_core.properties.property_manager import PropertyManager
from ds_capability import *
from ds_capability.components.commons import Commons

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class AbstractFeatureIntentTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_get_mask(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        tbl = pa.table([pa.array([1,2,1,4,None,-1,7]), pa.array(list('abbbcad'))], names=['A','B'])
        mask = tools._extract_mask(tbl.column('A'), condition=(1, 'equal', None))
        self.assertEqual(2, pc.sum(mask).as_py())
        self.assertEqual(1, mask.null_count)
        mask = tools._extract_mask(tbl.column('A'), condition=(1, 'equal', None), mask_null=True)
        self.assertEqual(2, pc.sum(mask).as_py())
        self.assertEqual(0, mask.null_count)
        # multiple
        condition = [(2, 'greater', 'or_'),(0, 'less', None)]
        mask = tools._extract_mask(tbl.column('A'), condition=condition)
        self.assertEqual(3, pc.sum(mask).as_py())
        self.assertEqual(1, mask.null_count)
        # wrong compare type
        with self.assertRaises(ValueError) as context:
            mask = tools._extract_mask(tbl.column('B'), condition=(1, 'equal', None))
        self.assertTrue("The operator 'equal' is not supported for data type 'string'." in str(context.exception))


    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def pm_view(capability: str, task: str, section: str = None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability, {}).get(task, {})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result


def tprint(t: pa.table, headers: [str, list] = None, d_type: [str, list] = None, regex: [str, list] = None):
    _ = Commons.filter_columns(t.slice(0, 10), headers=headers, d_types=d_type, regex=regex)
    print(Commons.table_report(_).to_string())


if __name__ == '__main__':
    unittest.main()



# To use a consistent encoding
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import re

here = path.abspath(path.dirname(__file__))

def read(*parts):
    filename = path.join(here, *parts)
    with open(filename, encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='discovery-capability',
    version=find_version('ds_capability', '__init__.py'),
    description='Data Science to production accelerator',
    long_description=read('README.rst'),
    url='https://github.com/gigas64/discovery-capability',
    author='Gigas64',
    author_email='gigas64@opengrass.net',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Adaptive Technologies',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='data pipeline, data preprocessing, data processing pipeline',
    packages=find_packages(exclude=['tests', 'guides', 'data', 'jupyter']),
    license='BSD',
    include_package_data=True,
    package_data={
        # If any package contains *.yaml or *.csv files, include them:
        '': ['*.yaml', '*.csv'],
    },
    python_requires='>=3.8',
    install_requires=[
        'discovery-core',
        'pyarrow',
        'pandas',
        'numpy',
        'matplotlib',
        'seaborn',
        'scikit-learn',
        'scipy',
        'requests',
    ],
    extras_require={},
    entry_points={},
    test_suite='tests',
)



import os

TOKEN_ENV_VAR_NAME = "CORTEX_TOKEN"
API_ENDPOINT_ENV_VAR_NAME = "CORTEX_API_ENDPOINT"


# -------------------------------------- Token Loaders --------------------------------------
def load_token_from_env():
    return os.getenv(TOKEN_ENV_VAR_NAME, None)


def load_token(env_resolution_order=None, token=None):
    """
    Returns the first token it can find based on the resolution order
    """
    env_resolution_order = env_resolution_order if isinstance(env_resolution_order, list) else ["args","env"]
    token_finders_per_env = {
        "args": (lambda: token),
        "env": load_token_from_env
    }
    for env in env_resolution_order:
        if not env in token_finders_per_env:
            raise Exception(f"Invalid Env: {env}")
        else:
            _token = token_finders_per_env[env]()
        if _token is not None:
            return _token
    # raise Exception(f"Could not find token in envs: {env_resolution_order}")
    return None

# ------------------------------------ Endpoint Loaders ------------------------------------


def load_endpoint_from_env():
    return API_ENDPOINT_ENV_VAR_NAME


def load_api_endpoint(env_resolution_order=None, endpoint=None):
    """
    Returns the first api endpoint it can extract from the enviroment.
    The order of in which the environments are searched is dictated by the env_resolution_order
    """
    env_resolution_order = env_resolution_order if isinstance(env_resolution_order, list) else ["args","env"]
    endpoint_finders_per_env = {
        "args": (lambda: endpoint),
        "env": load_endpoint_from_env
    }
    for env in env_resolution_order:
        if not env in endpoint_finders_per_env:
            raise Exception(f"Invalid Env: {env}")
        else:
            _endpoint = endpoint_finders_per_env[env]()
        if _endpoint is not None:
            return _endpoint
    # raise Exception(f"Could not find endpoint in envs: {env_resolution_order}")
    return None


from abc import abstractmethod
import pandas as pd
import pyarrow as pa
from ds_core.components.abstract_component import AbstractComponent
from ds_capability.components.commons import Commons
from ds_capability.components.discovery import DataDiscovery


# noinspection PyArgumentList
class AbstractCommonComponent(AbstractComponent):

    @classmethod
    @abstractmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, creator: str, uri_pm_repo: str=None, pm_file_type: str=None,
                 pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None, default_save=None,
                 reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, has_contract: bool=None):
        return cls

    def run_component_pipeline(self, intent_levels: [str, int, list]=None, run_book: str=None, seed: int=None,
                               reset_changed: bool=None, has_changed: bool=None, **kwargs):
        """runs the synthetic component pipeline. By passing an int value as the canonical will generate a synthetic
        file of that size

        :param intent_levels: (optional) a single or list of intent levels to run
        :param run_book: (optional) a saved runbook to run
        :param seed: (optional) a seed value for this run
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param kwargs: any additional kwargs
        """
        run_book = run_book if isinstance(run_book, str) and self.pm.has_run_book(run_book) else self.pm.PRIMARY_RUN_BOOK
        if isinstance(intent_levels, (str, int, list)):
            intent_levels = Commons.list_formatter(intent_levels)
        elif isinstance(run_book, str) and self.pm.has_run_book(book_name=run_book):
            intent_levels = self.pm.get_run_book(book_name=run_book)
        else:
            intent_levels = list(self.pm.get_intent().keys())
        canonical = None
        if self.pm.has_connector(self.CONNECTOR_SOURCE):
            canonical = self.load_source_canonical(reset_changed=reset_changed, has_changed=has_changed)
        for level in intent_levels:
            canonical = self.intent_model.run_intent_pipeline(canonical=canonical, intent_level=level, seed=seed, **kwargs)
        self.save_persist_canonical(canonical)
        return

    def load_source_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                              **kwargs) -> pa.Table:
        """returns the contracted source data as a DataFrame

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_SOURCE, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    def load_canonical(self, connector_name: str, reset_changed: bool=None, has_changed: bool=None,
                       return_empty: bool=None, **kwargs) -> pa.Table:
        """returns the canonical of the referenced connector

        :param connector_name: the name or label to identify and reference the connector
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        canonical = super().load_canonical(connector_name=connector_name, reset_changed=reset_changed,
                                           has_changed=has_changed, return_empty=return_empty, **kwargs)
        return canonical

    def load_persist_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                               **kwargs) -> pa.Table:
        """loads the clean pandas.DataFrame from the clean folder for this contract

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_PERSIST, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    def save_persist_canonical(self, canonical, auto_connectors: bool=None, **kwargs):
        """Saves the canonical to the clean files folder, auto creating the connector from template if not set"""
        if auto_connectors if isinstance(auto_connectors, bool) else True:
            if not self.pm.has_connector(self.CONNECTOR_PERSIST):
                self.set_persist()
        self.persist_canonical(connector_name=self.CONNECTOR_PERSIST, canonical=canonical, **kwargs)

    def add_column_description(self, column_name: str, description: str, save: bool=None):
        """ adds a description note that is included in with the 'report_column_catalog'"""
        if isinstance(description, str) and description:
            self.pm.set_intent_description(level=column_name, text=description)
            self.pm_persist(save)
        return

    @staticmethod
    def quality_report(canonical: pa.Table, nulls_threshold: float=None, dom_threshold: float=None,
                     cat_threshold: int=None, stylise: bool=None):
        """ Analyses a dataset, passed as a DataFrame and returns a quality summary

        :param canonical: The table to view.
        :param cat_threshold: (optional) The threshold for the max number of unique categories. Default is 60
        :param dom_threshold: (optional) The threshold limit of a dominant value. Default 0.98
        :param nulls_threshold: (optional) The threshold limit of a nulls value. Default 0.9
        :param stylise: (optional) if the output is stylised
        """
        stylise = stylise if isinstance(stylise, bool) else True
        return DataDiscovery.data_quality(canonical=canonical, nulls_threshold=nulls_threshold,
                                          dom_threshold=dom_threshold, cat_threshold=cat_threshold, stylise=stylise)
    @staticmethod
    def canonical_report(canonical: pa.Table, headers: [str,list]=None, regex:[str,list]=None, d_types:list=None,
                         drop: bool=None, stylise: bool=None, display_width: int=None):
        """The Canonical Report is a data dictionary of the canonical providing a reference view of the dataset's
        attribute properties

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param display_width: (optional) the width of the observational display
        """
        stylise = stylise if isinstance(stylise, bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types, drop=drop)
        return DataDiscovery.data_dictionary(canonical=tbl, stylise=stylise, display_width=display_width)

    @staticmethod
    def schema_report(canonical: pa.Table, headers: [str,list]=None, regex:[str,list]=None, d_types:list=None,
                      drop: bool=None, stylise: bool=True, table_cast: bool=None):
        """ presents the current canonical schema

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param table_cast: (optional) if the column should try to be cast to its type
        """
        stylise = stylise if isinstance(stylise, bool) else True
        table_cast = table_cast if isinstance(table_cast,bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types, drop=drop)
        return DataDiscovery.data_schema(canonical=tbl, table_cast=table_cast, stylise=stylise)

    @staticmethod
    def table_report(canonical: pa.Table, head: int=None):
        """ Creates a report from a pyarrow table in a tabular form

        :param canonical: the table to view
        :param head: The number of rows to show. Default to 20
        """
        head = head if isinstance(head, int) else 20
        return Commons.table_report(canonical, head=head)

    def report_task(self, stylise: bool=True):
        """ generates a report on the source contract

        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        report = self.pm.report_task_meta()
        df = pd.DataFrame.from_dict(data=report, orient='index').reset_index()
        df.columns = ['name', 'value']
        # sort out any values that start with a $ as it throws formatting
        for c in df.columns:
            df[c] = [f"{x[1:]}" if str(x).startswith('$') else x for x in df[c]]
        if stylise:
            return Commons.report(df, index_header='name')
        return pa.Table.from_pandas(df)

    def report_connectors(self, connector_filter: [str, list]=None, inc_pm: bool=None, inc_template: bool=None,
                          stylise: bool=True):
        """ generates a report on the source contract

        :param connector_filter: (optional) filters on the connector name.
        :param inc_pm: (optional) include the property manager connector
        :param inc_template: (optional) include the template connectors
        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        report = self.pm.report_connectors(connector_filter=connector_filter, inc_pm=inc_pm,
                                           inc_template=inc_template)
        df = pd.DataFrame.from_dict(data=report)
        # sort out any values that start with a $ as it throws formatting
        for c in df.columns:
            df[c] = [f"{x[1:]}" if str(x).startswith('$') else x for x in df[c]]
        if stylise:
            return Commons.report(df, index_header='connector_name')
        return pa.Table.from_pandas(df)

    def report_intent_description(self, action: [str, list]=None, stylise: bool=True):
        """ generates a report on the intent action notes

        :param action: (optional) filters on specific action names.
        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        df = pd.DataFrame.from_dict(data=self.pm.report_intent(levels=action, as_description=True,
                                                               level_label='column_name'))
        if stylise:
            return Commons.report(df, index_header='column_name')
        return pa.Table.from_pandas(df)

    def report_run_book(self, stylise: bool=True):
        """ generates a report on all the intent

        :param stylise: returns a stylised dataframe with formatting
        :return: pa.Table
        """
        df = pd.DataFrame.from_dict(data=self.pm.report_run_book())
        if stylise:
            return Commons.report(df, index_header='name')
        return pa.Table.from_pandas(df)

    def report_environ(self, stylise: bool=True):
        """ generates a report on all the intent

        :param stylise: returns a stylised dataframe with formatting
        :return: pa.Table
        """
        df = pd.DataFrame.from_dict(data=super().report_environ(), orient='index').reset_index()
        df.columns = ["environ", "value"]
        if stylise:
            return Commons.report(df, bold='environ')
        return pa.Table.from_pandas(df)

    def report_intent(self, levels: [str, int, list]=None, stylise: bool=True):
        """ generates a report on all the intent

        :param levels: (optional) a filter on the levels. passing a single value will report a single parameterised view
        :param stylise: (optional) returns a stylised dataframe with formatting
        :return: pa.Table
        """
        if isinstance(levels, (int, str)):
            df = pd.DataFrame.from_dict(data=self.pm.report_intent_params(level=levels))
            if stylise:
                return Commons.report(df, index_header='order')
        df = pd.DataFrame.from_dict(data=self.pm.report_intent(levels=levels))
        if stylise:
            return Commons.report(df, index_header='level')
        return pa.Table.from_pandas(df)

    def report_notes(self, catalog: [str, list]=None, labels: [str, list]=None, regex: [str, list]=None,
                     re_ignore_case: bool=False, stylise: bool=True, drop_dates: bool=False):
        """ generates a report on the notes

        :param catalog: (optional) the catalog to filter on
        :param labels: (optional) s label or list of labels to filter on
        :param regex: (optional) a regular expression on the notes
        :param re_ignore_case: (optional) if the regular expression should be case-sensitive
        :param stylise: (optional) returns a stylised dataframe with formatting
        :param drop_dates: (optional) excludes the 'date' column from the report
        :return: pa.Table
        """
        report = self.pm.report_notes(catalog=catalog, labels=labels, regex=regex, re_ignore_case=re_ignore_case,
                                      drop_dates=drop_dates)
        df = pd.DataFrame.from_dict(data=report)
        if stylise:
            return Commons.report(df, index_header='section', bold='label')
        return pa.Table.from_pandas(df)

    @classmethod
    def __dir__(cls):
        """returns the list of available methods associated with the parameterized intent"""
        rtn_list = []
        for m in dir(cls):
            if not m.startswith('_'):
                rtn_list.append(m)
        return rtn_list
